import asyncio
import base64
import hashlib
import hmac
import os
from functools import cached_property
from typing import Any, Callable, Coroutine, Dict, List, Optional, Tuple, Union

import aiohttp
from aiohttp import ClientResponse, ClientSession, ClientWebSocketResponse, WSMsgType
from aiohttp.client_exceptions import ContentTypeError
from aiohttp.formdata import FormData
from inflection import underscore

from fluid import json
from fluid.node import NodeWorker
from fluid.types import JsonType, String, to_bytes

from .utils import with_ops


class ResponseError(RuntimeError):
    pass


ObjectResponse = Union[Dict, "Entity"]
ListResponse = List[ObjectResponse]
ResponseType = Union[ListResponse, ObjectResponse, None]


class HttpResponseError(RuntimeError):
    def __init__(self, response: ClientResponse, data: Dict) -> None:
        self.response = response
        self.data = data
        self.data["request_url"] = str(response.url)
        self.data["request_method"] = response.method
        self.data["response_status"] = response.status

    @property
    def status(self) -> int:
        return self.response.status

    def __str__(self) -> str:
        return json.dumps(self.data, indent=4)


class HttpBase:
    @cached_property
    def extra_kwargs(self) -> Tuple[str, ...]:
        return ("wrap", "callback", "headers")

    async def get(self, url: str, **kwargs) -> ResponseType:
        kwargs["method"] = "GET"
        return await self.execute(url, **kwargs)

    async def patch(self, url: str, **kwargs) -> ResponseType:
        kwargs["method"] = "PATCH"
        return await self.execute(url, **kwargs)

    async def post(self, url: str, **kwargs) -> ResponseType:
        kwargs["method"] = "POST"
        return await self.execute(url, **kwargs)

    async def put(self, url: str, **kwargs) -> ResponseType:
        kwargs["method"] = "PUT"
        return await self.execute(url, **kwargs)

    async def delete(self, url: str, **kwargs) -> ResponseType:
        kwargs["method"] = "DELETE"
        return await self.execute(url, **kwargs)

    def nowrap(self, data, response: ClientResponse) -> ObjectResponse:
        return data

    def wrap(self, data, response: ClientResponse) -> ObjectResponse:
        return data

    def wrap_list(self, data, response: ClientResponse) -> ListResponse:
        return [self.wrap(d, response) for d in data]

    @classmethod
    def hmac(
        cls, key: String, message: String, alg: str = None, digest: str = "hex"
    ) -> String:
        sig = hmac.new(to_bytes(key), to_bytes(message), alg or "sha256")
        if digest == "hex":
            return sig.hexdigest()
        elif digest == "base64":
            return base64.b64encode(sig.digest())
        else:
            return sig.digest()

    @classmethod
    def hash(cls, data: String, alg: str = None, digest: str = "hex") -> String:
        h = hashlib.new(alg or "sha256", to_bytes(data))
        if digest == "hex":
            return h.hexdigest()
        elif digest == "base64":
            return base64.b64encode(h.digest())
        return h.digest()

    def kwargs(self, kwargs: Dict, key_name: str = None, **kw) -> Dict:
        """Expand a key arguments dictionary by extracting additional key valued
        parameters
        """
        if key_name and key_name not in kwargs:
            data = with_ops(kwargs, key_name)
            result = {key_name: data}
            result.update(
                ((key, data.pop(key)) for key in self.extra_kwargs if key in data)
            )
        else:
            result = kwargs.copy()
        result.update(kw)
        return result

    async def form_data(self, data: Dict) -> Coroutine[Any, Any, bytes]:
        return await FD(data).data()


class HttpClient(HttpBase):
    """Base class for Http clients"""

    session: ClientSession = None
    session_owner: bool = False
    user_agent: str = os.getenv("HTTP_USER_AGENT", "quantmind")
    content_type: str = "application/json"
    ResponseError: HttpResponseError = HttpResponseError
    ok_status = frozenset((200, 201))

    @property
    def cli(self) -> "HttpClient":
        return self

    def get_session(self) -> ClientSession:
        if not self.session:
            self.session_owner = True
            self.session = aiohttp.ClientSession()
        return self.session

    async def close(self) -> None:
        if self.session and self.session_owner:
            await self.session.close()
            self.session = None

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()

    async def execute(
        self,
        url: str,
        *,
        method: str = None,
        headers: Dict = None,
        callback=None,
        wrap=None,
        **kw,
    ) -> ResponseType:
        session = self.get_session()
        _headers = self.default_headers()
        _headers.update(headers or ())
        method = method or "GET"
        response = await session.request(method, url, headers=_headers, **kw)
        if callback:
            return await callback(response)
        if response.status in self.ok_status:
            data = await self.response_data(response)
        elif response.status == 204:
            return
        else:
            data = await self.response_error(response)
        wrap = wrap or self.wrap
        return wrap(data, response)

    def default_headers(self) -> Dict[str, str]:
        return {"user-agent": self.user_agent, "accept": self.content_type}

    def mock(self) -> None:
        pass

    @classmethod
    async def response_error(cls, response: ClientResponse) -> ResponseType:
        try:
            data = await cls.response_data(response)
        except ContentTypeError:
            data = dict(message=await response.text())
        raise cls.ResponseError(response, data)

    @classmethod
    async def response_data(cls, response):
        return await response.json()


class HttpComponent(HttpBase):
    @property
    def cli(self) -> HttpClient:
        return self.root.cli

    def get_session(self) -> ClientSession:
        return self.cli.get_session()

    async def execute(self, url, **params):
        params.setdefault("wrap", self.wrap)
        return await self.cli.execute(url, **params)


class Component(HttpComponent):
    def __init__(self, root: HttpClient, name: str = "") -> None:
        self.root = root
        self.path = name or underscore(type(self).__name__)

    def __repr__(self) -> str:
        return self.url

    def __str__(self) -> str:
        return self.url

    @cached_property
    def url(self) -> str:
        return f"{self.cli.url}/{self.path}"


class WsConnection(NodeWorker):
    """Worker for listening to stream of websocket messages"""

    def __init__(
        self,
        ws_component: "WsComponent",
        ws_url: str,
        ws_connection: ClientWebSocketResponse,
    ):
        super().__init__()
        self.ws_component = ws_component
        self.ws_url = ws_url
        self.ws_connection = ws_connection
        self.on_text_message = None
        self.on_binary_message = None
        ws_component.ws_connections[ws_url] = self

    async def write_json(self, data: JsonType) -> None:
        await self.write_str(json.dumps(data))

    async def write_str(self, msg: str) -> None:
        await self.ws_connection.send_str(msg)

    async def teardown(self) -> None:
        self.ws_component.ws_connections.pop(self.ws_url, None)
        if self.ws_component.on_disconnect:
            self.ws_component.on_disconnect(self)

    async def work(self) -> None:
        async for msg in self.ws_connection:
            if msg.type == WSMsgType.TEXT:
                if self.on_text_message:
                    try:
                        self.on_text_message(msg.data)
                    except Exception:
                        self.logger.exception(
                            "Critical failure while broadcasting text message"
                        )
            elif msg.type == WSMsgType.BINARY:
                if self.on_binary_message:
                    try:
                        self.on_binary_message(msg.data)
                    except Exception:
                        self.logger.exception(
                            "Critical failure while broadcasting binary message"
                        )
            # release the loop for greedy websocket connections
            await asyncio.sleep(0)


class WsComponent(Component):
    connection_factory = WsConnection
    on_disconnect: Optional[Callable[[WsConnection], None]] = None
    heartbeat: Optional[float] = None

    @cached_property
    def ws_connections(self) -> Dict[str, WsConnection]:
        return {}

    async def connect_and_listen(
        self,
        ws_url: str,
        on_text_message: Callable[[str], None] = None,
        on_binary_message: Callable[[bytes], None] = None,
        **kw,
    ) -> WsConnection:
        connection = self.ws_connections.get(ws_url)
        if not connection:
            if self.heartbeat:
                kw.setdefault("heartbeat", self.heartbeat)
            ws = await self.get_session().ws_connect(ws_url, **kw)
            connection = self.connection_factory(self, ws_url, ws)
            await connection.start()
        if on_text_message:
            connection.on_text_message = on_text_message
        if on_binary_message:
            connection.on_binary_message = on_binary_message
        return connection

    async def close(self) -> None:
        if self.ws_connections:
            ws = self.ws_connections.copy()
            self.ws_connections.clear()
            await asyncio.gather(*[c.close() for c in ws.values()])

    def wrap_json(self, on_message) -> Callable[[JsonType], None]:
        def on_text_message(msg: str) -> None:
            on_message(json.loads(msg))

        return on_text_message if on_message else on_message


class CrudMixin:
    async def get_list(self, **params) -> ListResponse:
        params.setdefault("wrap", self.wrap_list)
        return await self.execute(self.url, **self.kwargs(params, "params"))

    async def create_one(self, **kwargs) -> ObjectResponse:
        return await self.post(self.url, **self.kwargs(kwargs, "json"))

    create = create_one

    async def get_one(self, id_, **kwargs) -> ObjectResponse:
        return await self.get(f"{self.url}/{id_}", **self.kwargs(kwargs))

    async def update_one(self, id_, **kwargs) -> ObjectResponse:
        return await self.patch(f"{self.url}/{id_}", **self.kwargs(kwargs, "json"))

    async def delete_one(self, id_) -> None:
        return await self.delete(f"{self.url}/{id_}")


class Entity(HttpComponent):
    __slots__ = ("root", "data")

    def __init__(self, root: Component, data: Dict) -> None:
        self.root = root
        self.data = data

    def __repr__(self) -> str:
        return self.data.__repr__()

    def __str__(self) -> str:
        return self.data.__str__()

    def __getitem__(self, key: str) -> Any:
        return self.data[key]


class CRUDEntity(Entity):
    @property
    def id(self) -> str:
        return self.data.get("id")

    @property
    def url(self) -> str:
        return "%s/%s" % (self.root.url, self.id)


class FD:
    def __init__(self, data: Dict) -> None:
        fd = FormData(data)
        self.payload = fd()
        self.value = None

    async def data(self):
        await self.payload.write(self)
        return self.value

    async def write(self, value: bytes) -> None:
        self.value = value



from __future__ import annotations

import asyncio
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Optional

from .constants import FINISHED_STATES, TaskState

if TYPE_CHECKING:
    from .task import Task


@dataclass
class TaskRun:
    """A TaskRun contains all the data generated by a Task run"""

    id: str
    queued: int
    task: Task
    params: dict[str, Any]
    start: int = 0
    end: int = 0
    priority: str = ""
    state: str = TaskState.init.name
    waiter: asyncio.Future[Any] = field(default_factory=asyncio.Future)

    @property
    def in_queue(self) -> int:
        return self.start - self.queued

    @property
    def duration(self) -> int:
        return self.end - self.start

    @property
    def total(self) -> int:
        return self.end - self.queued

    @property
    def name(self) -> str:
        return self.task.name

    @property
    def name_id(self) -> str:
        return f"{self.task.name}.{self.id}"

    @property
    def exception(self) -> Optional[BaseException]:
        return self.waiter.exception() if self.waiter.done() else None

    @property
    def result(self) -> Any:
        return (
            self.waiter.result() if self.waiter.done() and not self.exception else None
        )

    @property
    def in_finish_state(self) -> bool:
        return TaskState[self.state] in FINISHED_STATES

    @property
    def is_failure(self) -> bool:
        return TaskState[self.state] is TaskState.failure

    def set_state(self, state: TaskState) -> None:
        self.state = state.name



from aiohttp.web import Application
from kong.client import Kong


def setup(app: Application) -> None:
    # kong client
    app["kong"] = Kong(request_kwargs=dict(ssl=False))
    app.on_shutdown.append(close_kong)


async def close_kong(app: Application) -> None:
    await app["kong"].close()



from .broker import Broker, QueuedTask
from .constants import TaskPriority, TaskState
from .consumer import TaskConsumer, TaskManager
from .crontab import Scheduler, crontab
from .every import every
from .scheduler import TaskScheduler
from .task import (
    Task,
    TaskConstructor,
    TaskContext,
    TaskDecoratorError,
    TaskExecutor,
    TaskRunError,
    create_task_app,
    task,
)
from .task_info import TaskInfo
from .task_run import TaskRun

__all__ = [
    "Scheduler",
    "TaskPriority",
    "TaskState",
    "TaskScheduler",
    "task",
    "Task",
    "TaskContext",
    "TaskRunError",
    "TaskExecutor",
    "TaskConstructor",
    "TaskDecoratorError",
    "TaskManager",
    "TaskConsumer",
    "TaskInfo",
    "TaskRun",
    "QueuedTask",
    "Broker",
    "create_task_app",
    "crontab",
    "every",
]



from __future__ import annotations

from abc import ABC, abstractmethod, abstractproperty
from functools import cached_property
from typing import Any, Dict, Iterable, List, NamedTuple, Optional
from uuid import uuid4

from yarl import URL

from fluid import json
from fluid.redis import FluidRedis
from fluid.utils import microseconds

from . import settings
from .constants import TaskPriority, TaskState
from .task import Task
from .task_info import TaskInfo
from .task_run import TaskRun

_brokers: dict[str, type[Broker]] = {}


def broker_url_from_env() -> URL:
    return URL(settings.SCHEDULER_BROKER_URL)


class TaskError(RuntimeError):
    pass


class UnknownTask(TaskError):
    pass


class DisabledTask(TaskError):
    pass


class TaskRegistry(Dict[str, Task]):
    def periodic(self) -> Iterable[Task]:
        for task in self.values():
            yield task


class QueuedTask(NamedTuple):
    run_id: str
    task: str
    params: Dict[str, Any]
    priority: Optional[TaskPriority] = None


class Broker(ABC):
    def __init__(self, url: URL) -> None:
        self.url: URL = url
        self.registry: TaskRegistry = TaskRegistry()

    @abstractproperty
    def task_queue_names(self) -> tuple[str, ...]:
        """Names of the task queues"""

    @abstractmethod
    async def queue_task(self, queued_task: QueuedTask) -> TaskRun:
        """Queue a task"""

    @abstractmethod
    async def get_task_run(self) -> Optional[TaskRun]:
        """Get a Task run from the task queue"""

    @abstractmethod
    async def queue_length(self) -> Dict[str, int]:
        """Length of task queues"""

    @abstractmethod
    async def get_tasks_info(self, *task_names: str) -> list[TaskInfo]:
        """List of TaskInfo objects"""

    @abstractmethod
    async def update_task(self, task: Task, params: dict[str, Any]) -> TaskInfo:
        """Update a task dynamic parameters"""

    async def close(self) -> None:
        """Close the broker on shutdown"""

    def new_uuid(self) -> str:
        return uuid4().hex

    async def filter_tasks(
        self, scheduled: Optional[bool] = None, enabled: Optional[bool] = None
    ) -> List[Task]:
        task_info = await self.get_tasks_info()
        task_map = {info.name: info for info in task_info}
        tasks = []
        for task in self.registry.values():
            if scheduled is not None and bool(task.schedule) is not scheduled:
                continue
            if enabled is not None and task_map[task.name].enabled is not enabled:
                continue
            tasks.append(task)
        return tasks

    def task_from_registry(self, task: str | Task) -> Task:
        if isinstance(task, Task):
            self.register_task(task)
            return task
        else:
            if task_ := self.registry.get(task):
                return task_
            raise UnknownTask(task)

    def register_task(self, task: Task) -> None:
        self.registry[task.name] = task

    async def enable_task(self, task_name: str, enable: bool = True) -> TaskInfo:
        """Enable or disable a registered task"""
        task = self.registry.get(task_name)
        if not task:
            raise UnknownTask(task_name)
        return await self.update_task(task, dict(enabled=enable))

    def task_run_from_data(self, data: Dict[str, Any]) -> TaskRun:
        """Build a TaskRun object from its metadata"""
        data = data.copy()
        name = data.pop("name")
        data["task"] = self.task_from_registry(name)
        return TaskRun(**data)

    def task_run_data(
        self, queued_task: QueuedTask, state: TaskState
    ) -> Dict[str, Any]:
        """Create a dictionary of metadata required by a task run

        This dictionary must be serializable by the broker
        """
        task = self.task_from_registry(queued_task.task)
        priority = queued_task.priority or task.priority
        return dict(
            id=queued_task.run_id,
            name=task.name,
            priority=priority.name,
            state=state.name,
            params=queued_task.params,
            queued=microseconds(),
        )

    @classmethod
    def from_url(cls, url: str = "") -> Broker:
        p = URL(url or broker_url_from_env())
        Factory = _brokers.get(p.scheme)
        if not Factory:
            raise RuntimeError(f"Invalid broker {p}")
        return Factory(p)

    @classmethod
    def register_broker(cls, name: str, factory: type[Broker]) -> None:
        _brokers[name] = factory


class RedisBroker(Broker):
    """A simple broker based on redis lists"""

    @cached_property
    def redis(self) -> FluidRedis:
        return FluidRedis(str(self.url.with_query({})), name=self.name)

    @property
    def name(self) -> str:
        return self.url.query.get("name", "redis-task-broker")

    @cached_property
    def task_queue_names(self) -> tuple[str, ...]:
        priorities = self.url.query.get("queues")
        if priorities is None:
            return tuple(self.task_queue_name(p) for p in TaskPriority)
        elif priorities:
            return tuple(
                self.task_queue_name(TaskPriority[p]) for p in priorities.split(",")
            )
        else:
            return ()

    def task_hash_name(self, name: str) -> str:
        return f"{self.name}-tasks-{name}"

    def task_queue_name(self, priority: TaskPriority) -> str:
        return f"{self.name}-queue-{priority.name}"

    async def get_tasks_info(self, *task_names: str) -> list[TaskInfo]:
        pipe = self.redis.cli.pipeline()
        names = task_names or self.registry
        requested_task_names = []
        for name in names:
            if name in self.registry:
                requested_task_names.append(name)
                pipe.hgetall(self.task_hash_name(name))
        tasks_info = await pipe.execute()
        return [
            self._decode_task(self.registry[name], task_info)
            for name, task_info in zip(requested_task_names, tasks_info)
        ]

    async def update_task(self, task: Task, params: dict[str, Any]) -> TaskInfo:
        pipe = self.redis.cli.pipeline()
        pipe.hset(
            self.task_hash_name(task.name),
            mapping={name: json.dumps(value) for name, value in params.items()},
        )
        pipe.hgetall(self.task_hash_name(task.name))
        _, info = await pipe.execute()
        return self._decode_task(task, info)

    async def queue_length(self) -> Dict[str, int]:
        if self.task_queue_names:
            pipe = self.redis.cli.pipeline()
            for name in self.task_queue_names:
                pipe.llen(name)
            result = await pipe.execute()
            return {p.name: r for p, r in zip(TaskPriority, result)}
        return {}

    async def close(self) -> None:
        """Close the broker on shutdown"""
        await self.redis.close()

    async def get_task_run(self) -> Optional[TaskRun]:
        if self.task_queue_names:
            data = await self.redis.cli.brpop(self.task_queue_names, timeout=1)
            if data:
                data_str = data[1].decode("utf-8")
                return self.task_run_from_data(json.loads(data_str))
        return None

    async def queue_task(self, queued_task: QueuedTask) -> TaskRun:
        task = self.task_from_registry(queued_task.task)
        priority = queued_task.priority or task.priority
        data = self.task_run_data(queued_task, TaskState.queued)
        await self.redis.cli.lpush(self.task_queue_name(priority), json.dumps(data))
        return self.task_run_from_data(data)

    def _decode_task(self, task: Task, data: dict[bytes, Any]) -> TaskInfo:
        info = {name.decode(): json.loads(value) for name, value in data.items()}
        return TaskInfo(
            name=task.name,
            description=task.description,
            schedule=str(task.schedule) if task.schedule else None,
            priority=task.priority.name,
            enabled=info.get("enabled", True),
            last_run_duration=info.get("last_run_duration"),
            last_run_end=info.get("last_run_end"),
            last_run_state=info.get("last_run_state"),
        )


Broker.register_broker("redis", RedisBroker)



from typing import Dict, Optional

from jsonschema import validate
from openapi.spec import SchemaParser


def validate_config(
    config: Dict, schema: Dict, defaults: bool = False
) -> Optional[Dict]:
    if schema:
        additional = schema.get("additionalProperties")
        schema["additionalProperties"] = True
        if defaults:
            config = _add_defaults(config, schema)
        validate(config, schema)
        if not additional:
            for key in tuple(config):
                if key not in schema["properties"]:
                    config.pop(key)
        return config


def json_meta_data(Dataclass: type, **kwargs) -> Dict:
    """Create a JSON payload with schema from a Dataclass"""
    parser = SchemaParser()
    parser.add_schema_to_parse(Dataclass)
    schemas = parser.parsed_schemas()
    schema = schemas.pop(Dataclass.__name__)
    if schemas:
        _replace_refs(schema, schemas)
    description = schema.pop("description")
    data = dict(description=description, schema=schema)
    data.update(kwargs)
    return data


# INTERNALS


def _add_defaults(config: Dict, schema: Dict) -> Dict:
    cfg = config.copy()
    for name, prop in schema["properties"].items():
        if name not in cfg and "default" in prop:
            cfg[name] = prop["default"]
    return cfg


def _replace_refs(schema, schemas):
    props = {}
    for name, field in schema["properties"].items():
        ref = field.get("$ref")
        is_array = False
        if not ref and field.get("type") == "array":
            is_array = True
            ref = field["items"].get("$ref")
        if ref and ref.startswith("#"):
            key = ref.split("/")[-1]
            if key in schemas:
                new_field = schemas[key]
                _replace_refs(schemas[key], schemas)
                if is_array:
                    field["items"] = new_field
                else:
                    description = field.get("description")
                    if description:
                        new_field["description"] = description
                    field = new_field
        props[name] = field
    schema["properties"] = props



from __future__ import annotations

import asyncio
import logging
from collections import defaultdict, deque
from dataclasses import dataclass
from functools import cached_property
from typing import Any, Callable, Coroutine, Deque, Dict, NamedTuple, Optional

from inflection import underscore

from fluid.node import Consumer, NodeWorkers, Worker
from fluid.utils import microseconds

from .broker import Broker, QueuedTask, TaskRegistry, UnknownTask
from .constants import TaskPriority, TaskState
from .task import Task
from .task_run import TaskRun

ConsumerCallback = Callable[[TaskRun, "TaskManager"], None]
AsyncExecutor = Callable[..., Coroutine[Any, Any, None]]
AsyncMessage = tuple[AsyncExecutor, tuple[Any, ...]]


class TaskFailure(RuntimeError):
    pass


@dataclass
class TaskManagerConfig:
    schedule_tasks: bool = True
    consume_tasks: bool = True
    max_concurrent_tasks: int = 10
    """number of coroutine workers"""
    sleep: float = 0.1
    """amount to sleep after completion of a task"""
    broker_url: str = ""


class Event(NamedTuple):
    type: str
    tag: str

    @classmethod
    def from_string(cls, event: str) -> "Event":
        bits = event.split(".")
        return cls(bits[0], bits[1] if len(bits) > 1 else "")


class TaskManager(NodeWorkers):
    """Base class for both TaskConsumer and TaskScheduler"""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__()
        self._msg_handlers: Dict[str, Dict[str, ConsumerCallback]] = defaultdict(dict)
        self._task_to_queue: Deque[QueuedTask] = deque()
        self._queue_tasks_worker = Worker(
            self._queue_tasks, logger=self.logger.getChild("queue")
        )
        self._consumer = Consumer(
            self._execute_async, logger=self.logger.getChild("internal-consumer")
        )
        self.config: TaskManagerConfig = TaskManagerConfig(**kwargs)
        self.add_workers(self._consumer)

    @cached_property
    def broker(self) -> Broker:
        return Broker.from_url(self.config.broker_url)

    @property
    def registry(self) -> TaskRegistry:
        return self.broker.registry

    @property
    def type(self) -> str:
        return underscore(self.__class__.__name__)

    async def setup(self) -> None:
        await self._queue_tasks_worker.start_app(self.app)

    async def teardown(self) -> None:
        await self._queue_tasks_worker.close()
        await self.broker.close()

    def register_task(self, task: Task) -> None:
        """Register a task with the task manager

        Only tasks registered can be executed by a task manager
        """
        self.broker.register_task(task)

    def queue(
        self,
        task: str,
        priority: Optional[TaskPriority] = None,
        **params: Any,
    ) -> QueuedTask:
        """Queue a Task for execution and return schedule_tasksthe QueuedTask object"""
        queued_task = QueuedTask(
            run_id=self.broker.new_uuid(),
            task=task,
            priority=priority,
            params=params,
        )
        self._task_to_queue.appendleft(queued_task)
        return queued_task

    def dispatch(self, task_run: TaskRun, event_type: str) -> None:
        """Dispatch a message to the registered handlers."""
        if handlers := self._msg_handlers.get(event_type):
            for handler in handlers.values():
                handler(task_run, self)

    def register_handler(self, event_name: str, handler: ConsumerCallback) -> None:
        event = Event.from_string(event_name)
        self._msg_handlers[event.type][event.tag] = handler

    def unregister_handler(self, event_name: str) -> None:
        event = Event.from_string(event_name)
        self._msg_handlers[event.type].pop(event.tag, None)

    def execute_async(self, async_callable: AsyncExecutor, *args: Any) -> None:
        self._consumer.submit((async_callable, args))

    # process tasks from the internal queue
    async def _queue_tasks(self) -> None:
        while self.is_running():
            try:
                queued_task = self._task_to_queue.pop()
            except IndexError:
                await asyncio.sleep(0.1)
            else:
                task_run = await self.broker.queue_task(queued_task)
                self.dispatch(task_run, "queued")
                await asyncio.sleep(0)

    async def _execute_async(self, message: AsyncMessage) -> None:
        try:
            executable, args = message
            await executable(*args)
        except Exception:
            self.logger.exception("unhandled exception while executing async callaback")


class TaskConsumer(TaskManager):
    """The Task Consumer is responsible for consuming tasks from a task queue"""

    def __init__(self, **config: Any) -> None:
        super().__init__(**config)
        self._concurrent_tasks: dict[str, dict[str, TaskRun]] = defaultdict(dict)
        self._priority_task_run_queue: deque[TaskRun] = deque()
        for i in range(self.config.max_concurrent_tasks):
            self.add_workers(
                Worker(
                    self._consume_tasks, logger=self.logger.getChild(f"worker-{i+1}")
                )
            )

    @property
    def num_concurrent_tasks(self) -> int:
        """The number of concurrent_tasks"""
        return sum(len(v) for v in self._concurrent_tasks.values())

    def num_concurrent_tasks_for(self, task_name: str) -> int:
        """The number of concurrent tasks for a given task_name"""
        return len(self._concurrent_tasks[task_name])

    async def queue_and_wait(self, task: str, **params: Any) -> Any:
        """Execute a task by-passing the broker task queue and wait for result"""
        return await self.execute(task, **params).waiter

    def execute(self, task: str, **params: Any) -> TaskRun:
        """Execute a Task by-passing the broker task queue"""
        queued_task = QueuedTask(
            run_id=self.broker.new_uuid(),
            task=task,
            params=params,
        )
        data = self.broker.task_run_data(queued_task, TaskState.queued)
        task_run = self.broker.task_run_from_data(data)
        self._priority_task_run_queue.appendleft(task_run)
        return task_run

    # Internals
    async def _consume_tasks(self) -> None:
        while self.is_running():
            if not self.config.consume_tasks:
                await asyncio.sleep(self.config.sleep)
                continue
            if self._priority_task_run_queue:
                task_run = self._priority_task_run_queue.pop()
            else:
                try:
                    maybe_task_run = await self.broker.get_task_run()
                except UnknownTask as exc:
                    self.logger.error(
                        "unknown task %s - it looks like it is not "
                        "registered with this consumer",
                        exc,
                    )
                    maybe_task_run = None
                if not maybe_task_run:
                    await asyncio.sleep(self.config.sleep)
                    continue
                else:
                    task_run = maybe_task_run
            task_name = task_run.name
            task_run.start = microseconds()
            task_run.set_state(TaskState.running)
            task_context = task_run.task.create_context(self, task_run=task_run)
            self._concurrent_tasks[task_name][task_run.id] = task_run
            #
            if task_run.task.max_concurrency < self.num_concurrent_tasks_for(task_name):
                task_run.set_state(TaskState.rate_limited)
                task_run.waiter.set_result(None)
            elif not (await self.broker.get_tasks_info(task_name))[0].enabled:
                task_run.set_state(TaskState.aborted)
                task_run.waiter.set_result(None)
            #
            else:
                task_context.logger.info("start")
                self.dispatch(task_run, "start")
                try:
                    result = await task_run.task.executor(task_context)
                except Exception as exc:
                    task_run.waiter.set_exception(exc)
                    task_run.set_state(TaskState.failure)
                else:
                    if task_run.is_failure:
                        task_run.waiter.set_exception(TaskFailure())
                    else:
                        task_run.waiter.set_result(result)
                        if not task_run.in_finish_state:
                            task_run.set_state(TaskState.success)
            try:
                await task_run.waiter
            except TaskFailure:
                # task_context.logger.warning("CPU bound task failure")
                # no need to log here, the log is already done from the CPU bound script
                pass
            except Exception:
                task_context.logger.exception("critical exception while executing")
            task_run.end = microseconds()
            self._concurrent_tasks[task_name].pop(task_run.id, None)
            await self.broker.update_task(
                task_run.task,
                dict(
                    last_run_end=task_run.end,
                    last_run_duration=task_run.duration,
                    last_run_state=task_run.state,
                ),
            )
            self.dispatch(task_run, "end")
            task_context.logger.log(
                logging.WARNING if task_run.is_failure else logging.INFO,
                "end - %s - milliseconds - %s",
                task_run.state,
                round(0.001 * task_run.duration, 3),
            )
            await asyncio.sleep(0)



import asyncio
from dataclasses import dataclass, field
from typing import Any

from .task_run import TaskRun


@dataclass
class WaitFor:
    run_id: str
    waiter: asyncio.Future[Any] = field(default_factory=asyncio.Future)

    def __call__(self, task_run: TaskRun, _: Any) -> None:
        if task_run.id == self.run_id:
            self.waiter.set_result(task_run)



import asyncio
import sys
from typing import Any, Callable, Dict, Optional

KernelCallback = Callable[[bytes], None]
READ_LIMIT = 2**16  # 64 KiB


async def run_python(*args: str) -> int:
    return await run(sys.executable, *args)


async def run(
    executable: str,
    *args: str,
    result_callback: Optional[KernelCallback] = None,
    error_callback: Optional[KernelCallback] = None,
    env: Optional[Dict[str, str]] = None,
    stream_output: bool = False,
    stream_error: bool = False,
) -> int:
    process = await asyncio.create_subprocess_exec(
        executable,
        *args,
        stdin=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        limit=READ_LIMIT,
        env=env,
    )
    if not process.stdout or not process.stderr:
        raise RuntimeError("Failed to create subprocess")
    await asyncio.wait(
        [
            asyncio.create_task(
                _read_stream(process.stdout, stream_output, result_callback or noop)
            ),
            asyncio.create_task(
                _read_stream(process.stderr, stream_error, error_callback or noop)
            ),
        ]
    )
    return await process.wait()


async def _read_line(stream: asyncio.StreamReader) -> bytes:
    chunks = []
    while True:
        try:
            chunk = await stream.readuntil()
        except asyncio.IncompleteReadError as e:
            chunks.append(e.partial)
            break
        except asyncio.LimitOverrunError:
            chunk = await stream.readexactly(READ_LIMIT)
            chunks.append(chunk)
        else:
            chunks.append(chunk)
            break
    return b"".join(chunks)


async def _read_stream(
    stream: asyncio.StreamReader, stream_output: bool, cb: KernelCallback
) -> None:
    while True:
        if stream_output:
            chunk = await _read_line(stream)
        else:
            chunk = await stream.read()
        if chunk:
            cb(chunk)
        else:
            break


def noop(data: bytes) -> None:
    pass


class EncodedLog:
    def __init__(self, log: Callable[[str], Any] = sys.stdout.write) -> None:
        self.log = log
        self.data: list[str] = []

    def __call__(self, data: bytes) -> None:
        msg = data.decode("utf-8")
        self.data.append(msg)
        self.log(msg)


class CollectBytes:
    data: bytes = b""

    def __call__(self, data: bytes) -> None:
        self.data += data



import os
from typing import Dict

from fluid.http import HttpClient

from .repo import GithubRepo


class Github(HttpClient):
    def __init__(self, token=None) -> None:
        self.token = token or get_token()

    @property
    def api_url(self):
        return "https://api.github.com"

    @property
    def uploads_url(self):
        return "https://uploads.github.com"

    def __repr__(self):
        return self.api_url

    __str__ = __repr__

    def repo(self, repo_name: str) -> GithubRepo:
        return GithubRepo(self, repo_name)

    def default_headers(self) -> Dict[str, str]:
        headers = super().default_headers()
        if self.token:
            headers["authorization"] = f"token {self.token}"
        return headers


def get_token() -> str:
    token = os.getenv("GITHUB_SECRET_TOKEN") or os.getenv("GITHUB_TOKEN")
    return token



import aiohttp_cors
from aiohttp.web import Application


def setup(app: Application):
    # Configure default CORS settings.
    cors = aiohttp_cors.setup(
        app,
        defaults={
            "*": aiohttp_cors.ResourceOptions(
                allow_credentials=True,
                allow_methods="*",
                expose_headers="*",
                allow_headers="*",
            )
        },
    )

    # Configure CORS on all routes.
    for route in list(app.router.routes()):
        try:
            cors.add(route, webview=True)
        except ValueError:
            pass



from dataclasses import dataclass
from typing import Any, Iterator, List, Type, Union

from aiohttp import web
from openapi import json
from openapi.data.validate import ValidationErrors
from openapi.db import CrudDB
from openapi.types import Record
from sqlalchemy import Table
from sqlalchemy.exc import NoResultFound
from sqlalchemy.ext.asyncio import AsyncResult

ErrorType = Union[Type[Exception], Exception]


class ExpectedOneOnly(ValueError):
    pass


@dataclass
class DbTools:
    db: CrudDB

    def raise_validation_error(self, message: str = "", errors=None) -> None:
        raw = self.as_errors(message, errors)
        data = self.dump(ValidationErrors, raw)
        raise web.HTTPUnprocessableEntity(
            body=json.dumps(data), content_type="application/json"
        )


async def batch_select(
    db: CrudDB,
    table: Table,
    *,
    batch_size: int = 50,
    **filters,
) -> Iterator[List[Record]]:
    query = db.get_query(table, table.select(), params=filters)
    async with db.transaction() as conn:
        result = await conn.stream(query)
        async for rows in result.partitions(batch_size):
            yield rows


def one_only(data: List, *, error: ErrorType = ExpectedOneOnly) -> Any:
    n = len(data)
    if n != 1:
        raise error
    return data[0]


def one_record(result: AsyncResult, *, error: ErrorType = ExpectedOneOnly) -> Record:
    try:
        return result.one()
    except NoResultFound as exc:
        raise error from exc



import logging
import os
from logging.config import dictConfig
from typing import Any, cast

import click

logger = logging.getLogger()


LOG_LEVEL = (os.environ.get("LOG_LEVEL") or "info").upper()
APP_NAME = os.environ.get("APP_NAME") or "fluid"
K8S = os.environ.get("KUBERNETES_SERVICE_HOST")
LOG_HANDLER = os.environ.get("LOG_HANDLER")


LOG_FORMAT_PRODUCTION = (
    "%(asctime)s %(name)s %(levelname)s %(filename)s %(funcName)s "
    "%(lineno)s %(module)s %(threadName)s %(message)s"
)

LOG_FORMAT_DEV = "%(asctime)s %(process)d %(levelname)s %(name)s %(message)s"

LOG_FORMAT = LOG_FORMAT_PRODUCTION if K8S else LOG_FORMAT_DEV


logger = logging.getLogger(APP_NAME)


def get_logger(name: str = "") -> logging.Logger:
    return logger.getChild(name) if name else logger


def level_num(level: str) -> int:
    return cast(int, getattr(logging, level))


def log_config(
    level: int,
    other_level: int = logging.WARNING,
    app_name: str | None = None,
    json: bool = bool(K8S),
) -> dict[str, Any]:
    app_name = app_name if app_name is not None else APP_NAME
    other_level = max(level, other_level)
    handler = LOG_HANDLER or ("main" if json else "color")
    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "json": {
                "()": "pythonjsonlogger.jsonlogger.JsonFormatter",
                "format": LOG_FORMAT,
            },
            "nicejson": {
                "()": "pythonjsonlogger.jsonlogger.JsonFormatter",
                "format": LOG_FORMAT,
                "json_indent": 2,
            },
            "color": {
                "()": "colorlog.ColoredFormatter",
                "format": f"%(log_color)s{LOG_FORMAT}",
            },
            "verbose": {"format": LOG_FORMAT},
        },
        "handlers": {
            "main": {
                "level": level,
                "class": "logging.StreamHandler",
                "formatter": "json",
            },
            "nicejson": {
                "level": level,
                "class": "logging.StreamHandler",
                "formatter": "nicejson",
            },
            "color": {
                "level": level,
                "class": "colorlog.StreamHandler",
                "formatter": "color",
            },
        },
        "loggers": {
            app_name: {"level": level, "handlers": [handler], "propagate": 0},
        },
        "root": {"level": other_level, "handlers": [handler]},
    }


def log_name(name: str = "") -> str:
    return APP_NAME if not name else f"{APP_NAME}.{name}"


def setup(level: str) -> None:
    dictConfig(log_config(level_num(level)))


@click.pass_context
def setup_logging(ctx: click.Context, verbose: bool, quiet: bool) -> None:
    if verbose:
        level = "DEBUG"
    elif quiet:
        level = "WARNING"
    else:
        level = LOG_LEVEL
    ctx.obj["log_level"] = level_num(level)
    setup(level)



import pycountry
from openapi.data import fields
from openapi.data.validate import ValidationError
from slugify import slugify

a2 = frozenset((c.alpha_2.lower() for c in pycountry.countries))


class SlugValidator(fields.StrValidator):
    def __call__(self, field, value):
        value = super().__call__(field, value)
        if slugify(value) != value:
            raise ValidationError(field.name, "Not a valid value")
        return value


class UUIDorSlugValidator(fields.UUIDValidator):
    def __call__(self, field, value):
        try:
            return super().__call__(field, value)
        except fields.ValidationError:
            if not isinstance(value, str):
                raise
            return value


def slug_field(max_length=None, min_length=None, **kw):
    kw.setdefault(
        "validator", SlugValidator(min_length=min_length, max_length=max_length)
    )
    return fields.data_field(**kw)


def uuid_field_or_slug(format="uuid", **kw):  # noqa
    """A UUID field with validation"""
    kw.setdefault("validator", UUIDorSlugValidator())
    return fields.data_field(format=format, **kw)


def country_field(**kw):
    kw["validator"] = fields.Choice(a2)
    return fields.str_field(**kw)



import enum


class TaskPriority(enum.Enum):
    high = 1
    medium = 2
    low = 3


class TaskState(enum.Enum):
    init = 0
    queued = 1
    running = 2
    success = 3
    failure = 4
    aborted = 5
    rate_limited = 6


FINISHED_STATES = frozenset(
    (TaskState.success, TaskState.failure, TaskState.aborted, TaskState.rate_limited)
)



import asyncio
import logging
import os
import sys
from logging.config import dictConfig
from typing import Any, overload

from fluid import kernel, log
from fluid.scheduler import (
    Scheduler,
    Task,
    TaskConstructor,
    TaskContext,
    TaskDecoratorError,
    TaskExecutor,
    TaskManager,
    TaskPriority,
    TaskRun,
    TaskState,
    create_task_app,
    settings,
)
from fluid.scheduler.task import RandomizeType


class RemoteLog:
    def __init__(self, out: Any) -> None:
        self.out = out

    def __call__(self, data: bytes) -> None:
        self.out.write(data.decode("utf-8"))


async def spawn(ctx: TaskContext) -> None:
    env = dict(os.environ)
    env["TASK_MANAGER_SPAWN"] = "true"
    result = await kernel.run(
        "python",
        "-W",
        "ignore",
        "-m",
        "fluid.scheduler.cpubound",
        ctx.name,
        ctx.task_run.id,
        result_callback=RemoteLog(sys.stdout),
        error_callback=RemoteLog(sys.stderr),
        env=env,
        stream_output=True,
        stream_error=True,
    )
    if result:
        ctx.task_run.set_state(TaskState.failure)


class CpuTaskConstructor(TaskConstructor):
    def __call__(self, executor: TaskExecutor) -> Task:
        if settings.TASK_MANAGER_SPAWN == "true":
            cpu_executor = executor
        else:
            self.kwargs["name"] = executor.__name__
            cpu_executor = spawn
        return super().__call__(cpu_executor)


@overload
def cpu_task(executor: TaskExecutor) -> Task:
    ...


@overload
def cpu_task(
    *,
    name: str | None = None,
    schedule: Scheduler | None = None,
    description: str | None = None,
    randomize: RandomizeType | None = None,
    max_concurrency: int = 1,
    priority: TaskPriority = TaskPriority.medium,
) -> CpuTaskConstructor:
    ...


def cpu_task(
    executor: TaskExecutor | None = None, **kwargs: Any
) -> Task | CpuTaskConstructor:
    """Decorator for creating a CPU bound task"""
    if kwargs and executor:
        raise TaskDecoratorError("cannot use positional parameters")
    elif kwargs:
        return CpuTaskConstructor(**kwargs)
    elif not executor:
        raise TaskDecoratorError("this is a decorator cannot be invoked in this way")
    else:
        return CpuTaskConstructor()(executor)


async def main(name: str, run_id: str) -> int:
    dictConfig(
        log.log_config(logging.INFO, logging.CRITICAL, f"{log.APP_NAME}.task.{name}")
    )
    app = create_task_app()
    task_manager: TaskManager = app["task_manager"]
    await app.startup()
    logger = task_manager.logger
    try:
        task = task_manager.broker.task_from_registry(name)
        task_run = TaskRun(id=run_id, queued=0, task=task, params={})
        task_context = task.create_context(task_manager, task_run=task_run)
        logger = task_context.logger
        await task.executor(task_context)
    except Exception:
        logger.exception("Unhandled exception executing CPU task")
        return 1
    finally:
        await app.shutdown()
    return 0


if __name__ == "__main__":
    exit(asyncio.get_event_loop().run_until_complete(main(*sys.argv[1:])))



import random
from datetime import datetime, timedelta

from .crontab import CronRun, Scheduler


class every(Scheduler):
    def __init__(
        self,
        delta: timedelta,
        delay: timedelta = timedelta(),
        jitter: timedelta = timedelta(),
    ) -> None:
        self.delta: timedelta = delta
        self.delay: timedelta = delay
        self.jitter: timedelta = jitter
        self._delta: timedelta = self.next_delta()
        self._started: datetime | None = None

    def info(self) -> str:
        return str(self.delta)

    def __call__(
        self, timestamp: datetime, last_run: CronRun | None = None
    ) -> CronRun | None:
        if not last_run and self.delay:
            if not self._started:
                self._started = timestamp
            if timestamp - self._started < self.delay:
                return None
        year, month, day, hour, minute, second, _, _, _ = timestamp.timetuple()
        run = CronRun(year, month, day, hour, minute, second)
        if not last_run or timestamp - last_run.datetime >= self._delta:
            self._delta = self.next_delta()
            return run
        return None

    def next_delta(self) -> timedelta:
        return self.delta + random.uniform(0, 1) * self.jitter



import asyncio
import uuid
from time import time
from typing import Any, Callable, Dict, Iterator, List, Optional, Sequence, Union, cast

import async_timeout
from inflection import underscore
from multidict import MultiDict
from openapi.spec.server import server_urls


def as_uuid(uid: Any) -> Optional[str]:
    if uid:
        if hasattr(uid, "hex"):
            uid = uid.hex
        try:
            return uuid.UUID(uid).hex
        except ValueError:
            return None
    return None


def add_api_url(
    request: Any, data: Union[List, Dict], path: Optional[str] = None, key: str = "id"
) -> Union[List, Dict]:
    """Add api_url field to data"""
    datas = data
    path = path if path is not None else request.path
    if not isinstance(data, list):
        datas = [cast(dict, data)]
    paths = (f"{path}/{d[key]}" for d in datas if key in d)
    for d, url in zip(datas, server_urls(request, paths)):
        d["api_url"] = url
    return data


def milliseconds() -> int:
    return int(time() * 1000)


def microseconds() -> int:
    return int(time() * 1000000)


def merge_dict(d1: Dict, d2: Dict) -> Dict:
    d1.update(d2)
    return d1


async def close_task(task: asyncio.Future, callback: Callable[[], None] = None) -> None:
    """Close an asyncio task cleanly"""
    if task and not task.done():
        try:
            task.cancel()
            await task
        except asyncio.CancelledError:
            pass
    if callback:
        await callback()
    return None


def dot_name(name: str) -> str:
    return ".".join(underscore(name).split("_"))


def query_dict(**data) -> MultiDict:
    return MultiDict(_query_dict(data))


def with_ops(data: Dict, key: str = "") -> Dict:
    data = {":".join(name.split("__")): value for name, value in data.items()}
    return MultiDict(multi(data)) if key == "params" else data


def multi(data: Dict) -> Iterator:
    for key, value in data.items():
        if isinstance(value, Sequence) and not isinstance(value, str):
            for inner_value in value:
                yield key, inner_value
        else:
            yield key, value


async def wait_for(assertion: Callable[[], bool], timeout=1) -> None:
    async with async_timeout.timeout(timeout):
        while True:
            if assertion():
                return
            await asyncio.sleep(0)
    assertion()


def _query_dict(data: Dict):
    for key, value in data.items():
        if isinstance(value, Sequence) and not isinstance(value, str):
            for d in value:
                yield key, d
        else:
            yield key, value



import atexit
import os
import signal
from collections import defaultdict
from contextlib import asynccontextmanager
from datetime import datetime
from tempfile import NamedTemporaryFile
from time import monotonic
from typing import Optional

import boto3
from aiohttp import web

from . import executor, kernel
from .node import Node

sampler_routes = web.RouteTableDef()


FLAMEGRAPH = os.getenv("FLAMEGRAPH_EXECUTABLE") or "/bin/flamegraph.pl"
FLAMEGRAPH_DATA_BUCKET, FLAMEGRAPH_DATA_PATH = os.getenv(
    "FLAMEGRAPH_DATA_BUCKET_PATH", "replace/me"
).split("/")


class FlamegraphError(RuntimeError):
    pass


class Sampler:
    """
    A simple stack sampler for low-overhead CPU profiling: samples the call
    stack every `interval` seconds and keeps track of counts by frame. Because
    this uses signals, it only works on the main thread.
    """

    def __init__(self, interval: float = 0.005) -> None:
        self.interval = interval
        self._started = None
        self._stack_counts = defaultdict(int)

    @property
    def started(self) -> float:
        return self._started

    def start(self) -> None:
        self.reset()
        signal.signal(signal.SIGVTALRM, self._sample)
        signal.setitimer(signal.ITIMER_VIRTUAL, self.interval)
        atexit.register(self.stop)

    def stop(self) -> None:
        self.reset()
        self._started = None
        signal.setitimer(signal.ITIMER_VIRTUAL, 0)

    def reset(self) -> None:
        self._started = monotonic()
        self._stack_counts = defaultdict(int)

    def stats(self) -> str:
        if self._started is None:
            return ""
        elapsed = monotonic() - self._started
        lines = ["elapsed {}".format(elapsed), "granularity {}".format(self.interval)]
        ordered_stacks = sorted(
            self._stack_counts.items(), key=lambda kv: kv[1], reverse=True
        )
        lines.extend(["{} {}".format(frame, count) for frame, count in ordered_stacks])
        return "\n".join(lines) + "\n"

    @asynccontextmanager
    async def flamegraph_file(self, title: str, stats: str) -> None:
        with NamedTemporaryFile(prefix="flamegraph-") as f:
            f.write(stats.encode("utf-8"))
            f.flush()
            result = kernel.CollectBytes()
            error = kernel.CollectBytes()
            code = await kernel.run(
                FLAMEGRAPH,
                f.name,
                "--title",
                title,
                result_callback=result,
                error_callback=error,
            )
            if code:
                raise FlamegraphError(error.data)
            yield result.data

    # INTERNALS

    def _sample(self, signum, frame) -> None:
        if self._started:
            stack = []
            while frame is not None:
                stack.append(self._format_frame(frame))
                frame = frame.f_back
            stack = ";".join(reversed(stack))
            self._stack_counts[stack] += 1
            signal.setitimer(signal.ITIMER_VIRTUAL, self.interval)

    def _format_frame(self, frame) -> str:
        return "{}({})".format(frame.f_code.co_name, frame.f_globals.get("__name__"))

    def __del__(self):
        self.stop()


class NodeSampler(Node):
    heartbeat = STACK_SAMPLER_PERIOD = int(os.getenv("STACK_SAMPLER_PERIOD", "60"))

    def __init__(self, title: str, sampler: Optional[Sampler] = None, **kwargs):
        super().__init__(**kwargs)
        self.title: str = title
        self.sampler: Sampler = sampler or Sampler()
        self.s3 = boto3.client("s3")

    async def tick(self) -> None:
        if not self.sampler.started:
            self.sampler.start()
        else:
            stats = self.sampler.stats()
            self.sampler.reset()
            try:
                async with self.sampler.flamegraph_file(self.title, stats) as svg:
                    await executor.run(self.upload, svg)
            except FlamegraphError as e:
                self.logger.warning("could not create flamegraph svg: %s", e)

    def upload(self, svg: bytes) -> None:
        dte = datetime.utcnow()
        date_str = dte.date().isoformat()
        time_str = dte.time().strftime("%H-%M-%S")
        s3_path = f"{FLAMEGRAPH_DATA_PATH}/{date_str}/{self.title}/{time_str}.svg"
        self.logger.info("upload flamegraph svg file to %s", s3_path)
        try:
            self.s3.put_object(
                Bucket=FLAMEGRAPH_DATA_BUCKET,
                Key=s3_path,
                Body=svg,
                ContentType="image/svg+xml",
            )
        except Exception:
            self.logger.exception(
                "Could not upload flamegraph data file to %s",
                s3_path,
            )


@sampler_routes.get("/stacksampler/stats")
async def stacksampler_stats(request):
    sampler: Sampler = request.app["service"].sampler
    return web.Response(text=sampler.stats())


@sampler_routes.get("/stacksampler/start")
async def stacksampler_start(request):
    sampler: Sampler = request.app["service"].sampler
    if not sampler.started:
        sampler.start()
        return web.Response(text="Sampler started")
    else:
        return web.Response(text="Sampler already started")


@sampler_routes.get("/stacksampler/stop")
async def stacksampler_stop(request):
    sampler: Sampler = request.app["service"].sampler
    if sampler.started:
        sampler.stop()
        return web.Response(text="Sampler stopped")
    else:
        return web.Response(text="Sampler already stopped")


@sampler_routes.get("/stacksampler/reset")
async def stacksampler_reset(request):
    sampler: Sampler = request.app["service"].sampler
    sampler.reset()
    return web.Response(text="Sampler reset")


@sampler_routes.get("/stacksampler/flamegraph")
async def stacksampler_flamegraph(request):
    sampler: Sampler = request.app["service"].sampler
    if sampler.started:
        title = request.app["cli"].name
        args = []
        with NamedTemporaryFile(prefix="flamegraph-") as f:
            f.write(sampler.stats().encode("utf-8"))
            f.flush()
            result = kernel.CollectBytes()
            error = kernel.CollectBytes()
            code = await kernel.run(
                FLAMEGRAPH,
                f.name,
                "--title",
                title,
                *args,
                result_callback=result,
                error_callback=error,
            )
            if code:
                return web.Response(body=error.data, content_type="text/plain")
            else:
                return web.Response(body=result.data, content_type="image/svg+xml")
    return web.Response(text="Sampler not started")



from __future__ import annotations

import inspect
import logging
import os
from importlib import import_module
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    NamedTuple,
    TypeVar,
    cast,
    overload,
)
from uuid import uuid4

from openapi.spec.utils import trim_docstring

from fluid import log
from fluid.node import WorkerApplication
from fluid.utils import microseconds

from .constants import TaskPriority
from .crontab import Scheduler
from .task_run import TaskRun

LogType = Callable[[str], None]

if TYPE_CHECKING:  # pragma: no cover
    from .broker import Broker
    from .consumer import TaskManager


class TaskDecoratorError(RuntimeError):
    pass


class TaskRunError(RuntimeError):
    pass


TASK_MANAGER_APP: str = os.getenv("TASK_MANAGER_APP", "")
T = TypeVar("T")


class ImproperlyConfigured(RuntimeError):
    pass


def create_task_app() -> WorkerApplication:
    if not TASK_MANAGER_APP:
        raise ImproperlyConfigured("missing TASK_MANAGER_APP environment variable")
    bits = TASK_MANAGER_APP.split(":")
    if len(bits) != 2:
        raise ImproperlyConfigured(
            "TASK_MANAGER_APP must be of the form <module>:<function>"
        )
    mod = import_module(bits[0])
    return cast(WorkerApplication, getattr(mod, bits[1])())


class TaskContext(NamedTuple):
    task_manager: "TaskManager"
    task_run: TaskRun
    log: LogType
    logger: logging.Logger

    @property
    def task(self) -> Task:
        return self.task_run.task

    @property
    def run_id(self) -> str:
        return self.task_run.id

    @property
    def name(self) -> str:
        return self.task_run.name

    @property
    def params(self) -> dict[str, Any]:
        return self.task_run.params

    def model(self, factory: type[T]) -> T:
        """Create a model instance from the task params"""
        return factory(**self.task_run.params)

    def raise_error(self, msg: str) -> None:
        raise TaskRunError(msg)


TaskExecutor = Callable[[TaskContext], Coroutine[Any, Any, Any]]
RandomizeType = Callable[[], float | int]


class Task(NamedTuple):
    """A Task execute any time it is invoked"""

    name: str
    executor: TaskExecutor
    logger: logging.Logger
    description: str = ""
    schedule: Scheduler | None = None
    randomize: RandomizeType | None = None
    max_concurrency: int = 1
    """how many tasks can run in each consumer concurrently"""
    priority: TaskPriority = TaskPriority.medium

    async def register(self, broker: "Broker") -> None:
        pass

    async def __call__(
        self,
        task_manager: TaskManager | None = None,
        **kwargs: Any,
    ) -> Any:
        if task_manager is None:
            task_manager = create_task_app()["task_manager"]
        context = self.create_context(task_manager, **kwargs)
        return await self.executor(context)

    def create_context(
        self,
        task_manager: TaskManager,
        task_run: TaskRun | None = None,
        log: LogType | None = None,
        **params: Any,
    ) -> TaskContext:
        if task_run is None:
            task_run = TaskRun(
                id=uuid4().hex, queued=microseconds(), task=self, params=params
            )
        else:
            assert task_run.name == self.name, "task_run for a different task"
            assert not params, "task run is provided - cannot pass params"
        return TaskContext(
            task_manager=task_manager,
            task_run=task_run,
            log=log or self.logger.info,
            logger=self.logger.getChild(task_run.id),
        )


@overload
def task(executor: TaskExecutor) -> Task:
    ...


@overload
def task(
    *,
    name: str | None = None,
    schedule: Scheduler | None = None,
    description: str | None = None,
    randomize: RandomizeType | None = None,
    max_concurrency: int = 1,
    priority: TaskPriority = TaskPriority.medium,
) -> TaskConstructor:
    ...


# implementation of the task decorator
def task(executor: TaskExecutor | None = None, **kwargs: Any) -> Task | TaskConstructor:
    if kwargs and executor:
        raise TaskDecoratorError("cannot use positional parameters")
    elif kwargs:
        return TaskConstructor(**kwargs)
    elif not executor:
        raise TaskDecoratorError("this is a decorator cannot be invoked in this way")
    else:
        return TaskConstructor()(executor)


class TaskConstructor:
    def __init__(self, **kwargs: Any) -> None:
        self.kwargs = kwargs

    def __call__(self, executor: TaskExecutor) -> Task:
        kwargs: dict[str, Any] = {
            "name": get_name(executor),
            "description": trim_docstring(inspect.getdoc(executor) or ""),
            "executor": executor,
        }
        kwargs.update(self.kwargs)
        name = kwargs["name"]
        kwargs["logger"] = log.get_logger(f"task.{name}")
        return Task(**kwargs)


def get_name(o: Any) -> str:
    if hasattr(o, "__name__"):
        return str(o.__name__)
    elif hasattr(o, "__class__"):
        return str(o.__class__.__name__)
    else:
        return str(o)



import asyncio
import os
from datetime import datetime
from typing import Dict

from ..node import Node
from .consumer import TaskManager
from .crontab import CronRun
from .task import Task


class TaskScheduler(TaskManager):
    """A task manager for scheduling tasks"""

    def __init__(self) -> None:
        super().__init__()
        self.add_workers(ScheduleTasks(self))


class ScheduleTasks(Node):
    heartbeat = float(os.getenv("TASK_SCHEDULER_HEARTBEAT", "0.1"))

    def __init__(self, task_manager: TaskScheduler) -> None:
        super().__init__()
        self.task_manager: TaskScheduler = task_manager
        self.last_run: Dict[str, CronRun] = {}

    async def tick(self) -> None:
        if not self.task_manager.config.schedule_tasks:
            return
        now = datetime.utcnow()
        periodic_tasks = await self.task_manager.broker.filter_tasks(
            scheduled=True, enabled=True
        )
        for task in periodic_tasks:
            if task.schedule:
                run = task.schedule(now, self.last_run.get(task.name))
                if run:
                    self.last_run[task.name] = run
                    from_now = task.randomize() if task.randomize else 0
                    if from_now:
                        asyncio.get_event_loop().call_later(from_now, self._queue, task)
                    else:
                        self._queue(task)

    def _queue(self, task: Task) -> None:
        self.task_manager.queue(task.name)



import asyncio
import os
from typing import Any, Callable, Optional, Sequence

import async_timeout
from redis.asyncio import BlockingConnectionPool, Redis
from redis.asyncio.client import PubSub

from . import json
from .log import APP_NAME, get_logger
from .node import NodeWorker

REDIS_DEFAULT_URL = os.getenv("REDIS_URL", "redis://localhost:6379")
DEFAULT_CACHE_TIMEOUT = int(os.getenv("DEFAULT_CACHE_TIMEOUT", 300))
REDIS_MAX_CONNECTIONS_DEFAULT = int(os.getenv("MAX_REDIS_CONNECTIONS", "5"))
CACHE_KEY_PREFIX = os.getenv("CACHE_KEY_PREFIX", "cache")

logger = get_logger("redis")


def setup(
    app, redis_url: str = "", name: str = APP_NAME, app_key="redis"
) -> "FluidRedis":
    redis = FluidRedis(url=redis_url, name=name)
    app[app_key] = redis
    app.on_shutdown.append(redis.close)
    return redis


class MessageReceiver(NodeWorker):
    """Subscribe to a list of channels"""

    def __init__(
        self,
        sub: PubSub,
        on_message: Callable[[str, str], None],
        channels: Sequence[str] = (),
        patterns: Sequence[str] = (),
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.sub: PubSub = sub
        self.on_message = on_message
        self.channels = tuple(channels)
        self.patterns = tuple(patterns)

    async def setup(self) -> None:
        if self.channels:
            await self.sub.subscribe(*self.channels)
        if self.patterns:
            await self.sub.psubscribe(*self.patterns)

    async def work(self) -> None:
        while self.is_running():
            try:
                async with async_timeout.timeout(1):
                    message = await self.sub.get_message(ignore_subscribe_messages=True)
                    if message is not None:
                        self.on_message(message)
                    await asyncio.sleep(0)
            except asyncio.TimeoutError:
                pass

    async def teardown(self) -> None:
        await self.sub.close()


class FluidRedis:
    def __init__(
        self,
        url: str = "",
        name: str = "",
        max_connections: int = 0,
    ) -> None:
        self.cli: Redis = Redis(
            connection_pool=BlockingConnectionPool.from_url(
                url or REDIS_DEFAULT_URL,
                max_connections=max_connections or REDIS_MAX_CONNECTIONS_DEFAULT,
                client_name=name or APP_NAME,
            )
        )

    async def close(self, app=None) -> None:
        await self.cli.close()
        await self.cli.connection_pool.disconnect()

    def receiver(
        self,
        on_message: Callable[[str, str], None],
        channels: Sequence[str] = (),
        patterns: Sequence[str] = (),
        **kwargs,
    ) -> MessageReceiver:
        return MessageReceiver(
            self.cli.pubsub(),
            on_message,
            channels,
            patterns,
            **kwargs,
        )

    # CACHE UTILITIES

    def cache_key(self, key: str, *args) -> str:
        return "-".join(str(arg) for arg in (CACHE_KEY_PREFIX, key, *args))

    async def from_cache(
        self,
        key: str,
        *args,
        expire: int = 0,
        loader: Optional[Callable] = None,
    ) -> Any:
        """Load JSON data from redis cache"""
        cache_key = self.cache_key(key, *args)
        data = await self.cli.get(cache_key)
        if not data and loader:
            expire = expire or DEFAULT_CACHE_TIMEOUT
            data = await loader(*args)
            if expire:
                await self.cli.set(cache_key, json.dumps(data), ex=expire)
        elif data is not None:
            data = json.loads(data)
        return data

    async def invalidate_cache(self, key: str, *args) -> None:
        await self.cli.delete(self.cache_key(key, *args))



from typing import Any, Coroutine, Dict, List, Union

from aiohttp.web import Response

String = Union[bytes, str]


def to_bytes(s: String) -> bytes:
    return s.encode("utf-8") if isinstance(s, str) else s


def to_string(s: String) -> str:
    return s.decode("utf-8") if isinstance(s, bytes) else s


AsyncResponse = Coroutine[Any, Any, Response]

StrCoroutine = Coroutine[Any, Any, str]

ListCoroutine = Coroutine[Any, Any, List]

DictCoroutine = Coroutine[Any, Any, Dict]

AnyCoroutine = Coroutine[Any, Any, Any]

JsonType = Union[bool, str, float, int, Dict, List]

StrDict = Dict[str, Any]



import asyncio
from typing import Any, Coroutine

import click
from openapi.exc import ImproperlyConfigured


class RunnerError(ImproperlyConfigured):
    pass


def sync_run(coro: Coroutine) -> Any:
    try:
        return asyncio.get_event_loop().run_until_complete(coro)
    except ImproperlyConfigured as e:
        click.secho(str(e), err=True, color="red")
        raise click.Abort from None



from aiohttp.web import Application
from openapi.cli import OpenApiClient

from . import log


def dev_app(cli: OpenApiClient, log_level="DEBUG") -> Application:
    log.setup(log_level)
    return cli.get_serve_app()



import json
import os
from typing import Any


def nice_env_str(space: int = 4, trim_length: int = 100):
    lt = max(len(k) for k in os.environ) + space
    values = []
    for key, value in os.environ.items():
        if len(value) > trim_length + 3:
            value = f"{value[:trim_length]}..."
        k = f"{key}:".ljust(lt)
        values.append(f"{k}{value}")
    return "\n".join(values)


def nice_json(data: Any) -> str:
    if not isinstance(data, str):
        return json.dumps(data, indent=4)
    return data



import os

from openapi import sentry
from openapi.middleware import json_error

PYTHON_ENV = os.environ.get("PYTHON_ENV", "production")
ENVIRONMENT = os.environ.get("ENVIRONMENT", PYTHON_ENV)
SENTRY_DSN = os.environ.get("SENTRY_DSN", "")


def add_error_middleware(app):
    app["env"] = PYTHON_ENV
    sm = sentry.middleware(app, SENTRY_DSN, ENVIRONMENT)
    app.middlewares.append(json_error())
    app.middlewares.append(sm)



import asyncio
import inspect
import logging
import os
import random
import time
import uuid
from abc import ABC, abstractmethod
from functools import cached_property, wraps
from logging import Logger
from typing import Any, Callable, Coroutine, List, Optional, Tuple

from aiohttp.client import ClientConnectionError, ClientConnectorError
from aiohttp.web import Application, GracefulExit

from .log import get_logger
from .utils import close_task, dot_name, underscore

WorkerType = Callable[[], Coroutine[Any, Any, None]]


class Id:
    @classmethod
    def name(cls) -> str:
        """My name"""
        return underscore(cls.__name__)

    @cached_property
    def uid(self) -> str:
        """My unique ID"""
        return uuid.uuid4().hex

    @classmethod
    def create_logger(cls, logger: Optional[logging.Logger] = None) -> logging.Logger:
        return logger or get_logger(dot_name(cls.name()))


class IdLog(Id):
    @cached_property
    def logger(self):
        return self.create_logger()


class NodeBase(ABC, Id):
    exit_lag: int = 1
    app: Optional[Application] = None

    async def start_app(self, app: Application | None = None) -> None:
        """Start application"""
        self.app = app
        await self.start()

    async def close_app(self, app: Application | None = None) -> None:
        await self.close()

    @abstractmethod
    def is_running(self) -> bool:
        """True if the Node is running"""

    @abstractmethod
    async def start(self) -> None:
        """called when the node worker has started"""
        pass

    @abstractmethod
    async def close(self) -> None:
        """called when the node worker closed"""
        pass

    async def setup(self) -> None:
        """Called by the :meth:`.start` method when the worker starts

        This can be optionally implemented by derived classes
        """
        pass

    async def teardown(self) -> None:
        """Called my :meth:`close` when the worker is stopping.

        This can be optionally implemented by derived classes
        """
        pass

    async def done(self) -> None:
        try:
            await self.teardown()
        except Exception:
            self.logger.exception("unhandled exception while tear down worker")

    async def system_exit(self) -> None:
        """Gracefully exiting the app if possible"""
        if self.is_running():
            await self.done()
        self.system_exit_sync()

    def system_exit_sync(self) -> None:
        """Exit the app"""
        self.logger.warning("bailing out!")
        asyncio.get_event_loop().call_later(self.exit_lag, self._exit)

    def _exit(self) -> None:  # pragma: no cover
        if os.getenv("PYTHON_ENV") != "test":
            raise GracefulExit


class NodeWorker(NodeBase):
    def __init__(self, *, logger: Optional[Logger] = None) -> None:
        self.logger: Logger = self.create_logger(logger)
        self._worker = None

    @property
    def debug(self) -> bool:
        return self.logger.isEnabledFor(logging.DEBUG)

    # FOR DERIVED CLASSES

    async def work(self) -> None:
        """Main work coroutine, this is where you define the asynchronous loop.

        Must be implemented by derived classes
        """
        raise NotImplementedError

    # API

    def is_running(self) -> bool:
        """True if the Node is running"""
        return bool(self._worker)

    async def start(self) -> None:
        """Start the node"""
        assert not self.is_running(), "Node already running - cannot start"
        await self.setup()
        self._worker = asyncio.ensure_future(self._work())

    async def close(self, close_worker: bool = True) -> None:
        if self._worker:
            self.logger.info("closing")
            worker = self._worker
            self._worker = None
            if close_worker:
                await close_task(worker, self.done)
            else:
                await self.done()
            self.logger.warning("closed")

    # INTERNAL

    async def _work(self) -> None:
        self.logger.warning("started")
        try:
            await self.work()
        except asyncio.CancelledError:
            pass
        except Exception:
            self.logger.exception("unhandled exception in worker")
            await self.system_exit()
        else:
            await self.close(close_worker=False)


class WorkerApplication(dict[str, Any]):
    def __init__(self) -> None:
        super().__init__()
        self.on_startup = []
        self.on_shutdown = []

    async def startup(self) -> None:
        for on_startup in self.on_startup:
            await on_startup(self)

    async def shutdown(self) -> None:
        for on_shutdown in self.on_shutdown:
            await on_shutdown(self)


class NodeWorkers(NodeBase):
    def __init__(self, *workers: NodeWorker, logger: Optional[Logger] = None) -> None:
        self.logger: Logger = self.create_logger(logger)
        self._closing: bool = False
        self._workers: List[NodeBase] = list(workers)

    @property
    def debug(self) -> bool:
        return self.logger.isEnabledFor(logging.DEBUG)

    def is_running(self) -> bool:
        return isinstance(self._workers, tuple)

    def is_closing(self) -> bool:
        return self._closing

    def add_workers(self, *workers: NodeBase) -> None:
        if self.is_running():
            raise RuntimeError("Cannot add workers when started")
        self._workers.extend(workers)

    async def start(self) -> None:
        await self.setup()
        self.logger.warning("started")
        workers = self._freeze_workers()
        await asyncio.gather(*[w.start_app(self.app) for w in workers])

    async def close(self) -> None:
        if self.is_running():
            self._closing = True
            await asyncio.gather(*[w.close_app(self.app) for w in self._workers])
            await self.teardown()

    def _freeze_workers(self) -> Tuple[NodeBase, ...]:
        if isinstance(self._workers, tuple):
            raise RuntimeError("worker already started")
        self._workers = tuple(self._workers)
        return self._workers


class Node(NodeWorker):
    """A nodeworker with an heartbeat work loop and ability to publish
    messages into a pubsub
    """

    heartbeat: float = 1
    ticks: int = 0

    async def tick(self) -> None:
        """called at every iteration in the worker"""
        pass

    async def work(self) -> None:
        while True:
            start = time.monotonic()
            self.ticks += 1
            await self.tick()
            dt = time.monotonic() - start
            await asyncio.sleep(max(self.heartbeat - dt, 0))


class Consumer(NodeWorker):
    def __init__(
        self,
        process_message,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.process_message = process_message
        self._message_queue: Optional[asyncio.Queue] = None

    def qsize(self) -> int:
        return 0 if self._message_queue is None else self._message_queue.qsize()

    async def setup(self) -> None:
        self._message_queue = asyncio.Queue()

    async def work(self):
        while self.is_running():
            message = await self._message_queue.get()
            await self.process_message(message)
            await asyncio.sleep(0)

    def submit(self, message) -> None:
        if self._message_queue is None:
            raise RuntimeError("cannot submit to a non running consumer")
        self._message_queue.put_nowait(message)


class Worker(NodeWorker):
    def __init__(
        self,
        work: WorkerType,
        logger: Optional[Logger] = None,
    ) -> None:
        super().__init__(logger=logger)
        self.work = work


class TickWorker(Node):
    def __init__(
        self,
        tick: WorkerType,
        heartbeat: float = 1,
        logger: Optional[Logger] = None,
    ) -> None:
        super().__init__(logger=logger)
        self.heartbeat = heartbeat
        self.tick = tick


class every:
    def __init__(self, seconds: float, noise: float = 0) -> None:
        self.seconds = seconds
        self.noise = min(noise, seconds)
        self.last = 0
        self.gap = self._gap()
        self.ticks = 0

    def __call__(self, method):
        method.every = self

        @wraps(method)
        async def _(node, *args) -> None:
            now = time.time()
            if now - self.last > self.gap:
                self.last = now
                self.gap = self._gap()
                self.ticks += 1
                try:
                    await method(node, *args)
                except (ClientConnectionError, ClientConnectorError) as exc:
                    node.logger.error(str(exc))

        return _

    def _gap(self) -> float:
        return self.seconds + self.noise * (random.random() - 0.5)


def on_error_exit(
    method: Callable[[NodeBase, Any], None]
) -> Callable[[NodeBase, Any], None]:
    @wraps(method)
    def sync_wrap(node: NodeBase, *args) -> None:
        try:
            method(node, *args)
        except Exception:
            node.logger.exception("unhandled exception, bailing out!")
            node.system_exit_sync()

    @wraps(method)
    async def async_wrap(node: NodeBase, *args) -> None:
        try:
            await method(node, *args)
        except Exception:
            node.logger.exception("unhandled exception, bailing out!")
            await node.system_exit()

    return async_wrap if inspect.iscoroutinefunction(method) else sync_wrap



from pydantic import BaseModel, Field


class TaskInfo(BaseModel):
    name: str = Field(description="Task name")
    description: str = Field(description="Task description")
    priority: str = Field(description="Task priority")
    schedule: str | None = Field(default=None, description="Task schedule")
    enabled: bool = Field(default=True, description="Task enabled")
    last_run_end: int | None = Field(
        default=None, description="Task last run end as milliseconds since epoch"
    )
    last_run_duration: int | None = Field(
        default=None, description="Task last run duration in milliseconds"
    )
    last_run_state: str | None = Field(
        default=None, description="State of last task run"
    )



from base64 import b64encode
from typing import TYPE_CHECKING, Dict, List

from nacl import encoding, public

if TYPE_CHECKING:
    from .client import Github


class GithubRepo:
    def __init__(self, cli: "Github", repo_name: str) -> None:
        self.cli: "Github" = cli
        self.repo_name: str = repo_name
        self.branches: GithubBranches = GithubBranches(self, "branches")
        self.deployments: GithubDeployments = GithubDeployments(self, "deployments")
        self.secrets: GithubSecrets = GithubSecrets(self, "actions/secrets")

    @property
    def url(self) -> str:
        return f"{self.cli.api_url}/repos/{self.repo_name}"

    def __repr__(self) -> str:
        return self.url

    __str__ = __repr__

    async def info(self) -> Dict:
        return await self.cli.get(self.url)

    async def flat_info(self, raw: bool = False) -> Dict:
        info = await self.info()
        branch = await self.branches.get(info["default_branch"])
        if raw:
            info["default_branch"] = branch
            return info
        return {
            "name": info["full_name"],
            "delete branch on merge": info["delete_branch_on_merge"],
            "default branch": branch["name"],
            "default protected": branch["protected"],
            "default branch sha": branch["commit"]["sha"],
            "default branch date": branch["commit"]["commit"]["author"]["date"],
        }


class GithubRepoComponent:
    def __init__(self, repo: GithubRepo, path: str) -> None:
        self.repo: GithubRepo = repo
        self.path: str = path

    @property
    def url(self) -> str:
        return f"{self.repo.url}/{self.path}"

    @property
    def cli(self) -> "Github":
        return self.repo.cli

    def __repr__(self):
        return self.url

    __str__ = __repr__

    async def get_list(self) -> List[Dict]:
        return await self.cli.get(self.url)


class GithubBranches(GithubRepoComponent):
    async def get(self, branch: str) -> Dict:
        """Branch data"""
        return await self.cli.get(f"{self.url}/{branch}")


class GithubDeployments(GithubRepoComponent):
    special_accept = dict(
        in_progress="application/vnd.github.flash-preview+json",
        queued="application/vnd.github.flash-preview+json",
        inactive="application/vnd.github.ant-man-preview+json",
    )

    async def get(self, deployment_id: str) -> Dict:
        """Get a existing deployment"""
        return await self.cli.get(f"{self.url}/{deployment_id}")

    async def create(self, ref: str, auto_merge: bool = False, **params) -> Dict:
        """Create a new deployment"""
        return await self.cli.post(
            self.url,
            json=dict(ref=ref, auto_merge=auto_merge, **params),
        )

    async def update(self, deployment_id: str, **params) -> Dict:
        accept = self.special_accept.get(params.get("state"))
        headers = {}
        if accept:
            headers["accept"] = accept
        return await self.cli.post(
            f"{self.url}/{deployment_id}/statuses", json=params, headers=headers
        )


class GithubSecrets(GithubRepoComponent):
    public_key: Dict[str, str] = ""

    async def update(self, name: str, value: str) -> Dict:
        """Branch data"""
        public_key = await self.get_public_key()
        encrypted_value = encrypt(public_key["key"], value)
        return await self.cli.put(
            f"{self.url}/{name}",
            json=dict(encrypted_value=encrypted_value, key_id=public_key["key_id"]),
        )

    async def get_public_key(self) -> str:
        if not self.public_key:
            self.public_key = await self.cli.get(f"{self.url}/public-key")
        return self.public_key


def encrypt(public_key: str, secret_value: str) -> str:
    """Encrypt a Unicode string using the public key."""
    public_key = public.PublicKey(public_key.encode("utf-8"), encoding.Base64Encoder())
    sealed_box = public.SealedBox(public_key)
    encrypted = sealed_box.encrypt(secret_value.encode("utf-8"))
    return b64encode(encrypted).decode("utf-8")



import logging
import os
from typing import Callable, Optional

from aiohttp.web import Application
from openapi.middleware import json_error, sentry_middleware
from prometheus_async import aio

from . import backdoor, stacksampler, status

logger = logging.getLogger("fluid.service")


class Service:
    def __init__(self, env: str, metrics_path: str = "") -> None:
        self.env = env
        self.sampler = stacksampler.Sampler()

    async def on_startup(self, app: Application) -> None:
        pass

    async def on_shutdown(self, app: Application) -> None:
        pass

    @classmethod
    def setup(
        cls,
        app: Application = None,
        sentry_dsn: str = "",
        backdoor_port: int = 0,
        service_status: Optional[Callable] = None,
        status_routes: bool = True,
        metrics_path: str = "/metrics",
        sampler: bool = False,
    ) -> Application:
        """Setup boilerplate for asyncio services"""
        if app is None:
            app = Application()
        # create service object
        service = cls(app.get("env") or os.getenv("PYTHON_ENV", "prod"))
        #
        # Add metrics
        if metrics_path:
            app.router.add_get(metrics_path, aio.web.server_stats)
        #
        app["service"] = service
        if service_status:
            app["service_status"] = service_status
        app.on_startup.append(service.on_startup)
        app.on_shutdown.insert(0, service.on_shutdown)
        # add app health probes
        if status_routes:
            app.router.add_routes(status.status_routes)
        # add error handlers
        app.middlewares.append(json_error(status_codes=set(range(400, 504))))
        # add sentry handler
        if sentry_dsn:
            sentry_middleware(app, sentry_dsn, service.env)
        # add backdoor
        if backdoor_port:
            backdoor.setup(app, backdoor_port)

        if sampler:
            app.router.add_routes(stacksampler.sampler_routes)
        return app


setup = Service.setup



import os

from aiohttp import web
from openapi.spec.server import server_urls
from openapi.tz import utcnow

__version__ = os.environ.get("GIT_SHA", "unknown")
__timestamp__ = os.environ.get("TIMESTAMP", utcnow().isoformat())


status_routes = web.RouteTableDef()


@status_routes.get("/status")
async def status(request):
    service_status = request.app.get("service_status")
    if service_status:
        await service_status(request)
    result = dict(ok=True, sha=__version__, timestamp=__timestamp__)
    return web.json_response(result)


@status_routes.get("/headers")
async def headers(request):
    result = dict(request.headers.items())
    return web.json_response(result)


@status_routes.get("/")
async def urls(request):
    paths = set()
    for route in request.app.router.routes():
        route_info = route.get_info()
        path = route_info.get("path", route_info.get("formatter", None))
        paths.add(path)
    return web.json_response(server_urls(request, sorted(paths)))



from typing import Dict, List

from aiohttp import web

routes = web.RouteTableDef()


REDIS_CLIENT = [
    "id",
    "addR",
    "fd",
    "name",
    "age",
    "idle",
    "flags",
    "db",
    "sub",
    "psub",
    "multi",
    "qbuf",
    "qbuf-free",
    "obl",
    "oll",
    "omem",
    "events",
    "cmd",
]


def redis(request):
    return request.app["redis"].pub


@routes.get("/redis/clients")
async def status(request):
    data = await redis(request).client_list()
    return web.json_response([redis_client(d) for d in data])


def redis_client(d: List) -> Dict:
    return dict(zip(REDIS_CLIENT, d))
